/*****************************************************************************
 Feature Vectors define templated classes for 
 a) feature vector - CDimFeatureVector
 b) vector of feature vectors - CDimVectorOfFeatures

 DEFINITION

 Author: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
 Copyright (c) 2006 Vision Research Lab, UCSB <http://vision.ece.ucsb.edu>

 History:
   04/19/2006 16:20 - First creation

 Ver : 2
*****************************************************************************/

#ifndef DIM_FEATUREVECTORS
#define DIM_FEATUREVECTORS

#include <xvectors.h>
#include <fstream>

// define this macro not to use the STL containers for feature vectors
//#define DIM_NOSTL_CONTAINER_VECTOR


//----------------------------------------------------------------------------
// Feature Vector 
//----------------------------------------------------------------------------

#ifndef DIM_NOSTL_CONTAINER_VECTOR
//#pragma message(">>>>>  Feature vectors: STL based")

template <class T> 
class CDimFeatureVector: public std::vector<T>, public CDimListIO
{
public:
  //typedef typename std::vector<T>::iterator iterator;

  CDimFeatureVector( )
    : std::vector<T> ( (std::vector<T>::size_type) 1) { };
  CDimFeatureVector( unsigned int dimensions )
    : std::vector<T> ( (std::vector<T>::size_type) dimensions) { };
  CDimFeatureVector( unsigned int dimensions, T v )
    : std::vector<T> ( (std::vector<T>::size_type) dimensions, v) { };
  
  virtual ~CDimFeatureVector() { clear(); };
  virtual CDimFeatureVector* clone() const  { return new CDimFeatureVector(*this); };

  inline const T* data() { return *begin(); };
  void fill( T v );
  
  using std::vector<T>::assign;
  void assign( const T *v, int size );
  void assign( const CDimFeatureVector &v );

  // I/O partial
  using CDimListIO::fromFile;
  using CDimListIO::toFile;
  int fromFile(std::ifstream *fs);
  int toFile(std::ofstream *fs);

  // element-wise operators
  CDimFeatureVector<T>& operator+=(const CDimFeatureVector<T>& rv);
  CDimFeatureVector<T>& operator-=(const CDimFeatureVector<T>& rv);
  CDimFeatureVector<T>& operator*=(const CDimFeatureVector<T>& rv);
  CDimFeatureVector<T>& operator/=(const CDimFeatureVector<T>& rv);

  // operators with scalars
  CDimFeatureVector<T>& operator+=(double rv);
  CDimFeatureVector<T>& operator-=(double rv);
  CDimFeatureVector<T>& operator*=(double rv);
  CDimFeatureVector<T>& operator/=(double rv);

  // shifts
  void shiftRight ( unsigned int first, unsigned int num );
  void shiftLeft  ( unsigned int first, unsigned int num );
  void shiftRight ( ) { shiftRight ( 0, (unsigned int) size() ); }
  void shiftLeft  ( ) { shiftLeft  ( 0, (unsigned int) size() ); }

  // simple stdout
  void printf();
};

template<class T>
void CDimFeatureVector<T>::fill( T v ) { 
  std::vector<T>::iterator it = begin();
  while ( it<end() ) {
    *it = v;
    ++it;
  }
}

template<class T>
void CDimFeatureVector<T>::assign( const T *v, int size ) {
  if (v == NULL) return;
  resize( size );
  std::vector<T>::iterator it = begin();
  unsigned int i=0;
  while ( it<end() ) {
    *it = v[i];
    ++it;  ++i;
  }
}

template<class T>
void CDimFeatureVector<T>::assign( const CDimFeatureVector &v ) {
  resize( v.size() );
  std::vector<T>::iterator it = begin();
  unsigned int i=0;
  while ( it<end() ) {
    *it = v[i];
    ++it;  ++i;
  }
}

template<class T>
int CDimFeatureVector<T>::fromFile(std::ifstream *fs) {
  if (fs == NULL) return -1;
  for (unsigned int i=0; i<this->size(); ++i) {
    T v; 
    fs->read((char*)&v, sizeof(T));
    if (fs->fail()) return -1; 
    (*this)[i] = v;
  }
  return (int) this->size() * sizeof(T);
}

template<class T>
int CDimFeatureVector<T>::toFile(std::ofstream *fs) {
  if (fs == NULL) return -1;
  for (unsigned int i=0; i<this->size(); ++i) {
    T v = (*this)[i];
    fs->write((char*)&v, sizeof(T));
    if (fs->fail()) return -1; 
  }
  return (int) this->size() * sizeof(T);
}

//--------------------------------
// element-wise operators
//--------------------------------

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator+=(const CDimFeatureVector<T>& rv) {
  unsigned int i=0;
  unsigned int sz = (size() < rv.size()) ? size() : rv.size();

  iterator it = begin();
  while ( (i<sz) && (it<end()) ) {
    *it += rv[i];
    ++it;
    ++i;
  }
  return *this;
}

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator-=(const CDimFeatureVector<T>& rv) {
  unsigned int i=0;
  unsigned int sz = (size() < rv.size()) ? size() : rv.size();

  iterator it = begin();
  while ( (i<sz) && (it<end()) ) {
    *it -= rv[i];
    ++it;
    ++i;
  }
  return *this;
}

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator*=(const CDimFeatureVector<T>& rv) {
  unsigned int i=0;
  unsigned int sz = (size() < rv.size()) ? size() : rv.size();

  iterator it = begin();
  while ( (i<sz) && (it<end()) ) {
    *it *= rv[i];
    ++it;
    ++i;
  }
  return *this;
}

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator/=(const CDimFeatureVector<T>& rv) {
  unsigned int i=0;
  unsigned int sz = (size() < rv.size()) ? size() : rv.size();

  iterator it = begin();
  while ( (i<sz) && (it<end()) ) {
    *it /= rv[i];
    ++it;
    ++i;
  }
  return *this;
}

//--------------------------------
// operators with scalars
//--------------------------------

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator+=(double rv) {
  iterator it = begin();
  while (it<end()) {
    *it += rv;
    ++it;
  }
  return *this;
}

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator-=(double rv) {
  iterator it = begin();
  while (it<end()) {
    *it -= rv;
    ++it;
  }
  return *this;
}

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator*=(double rv) {
  iterator it = begin();
  while (it<end()) {
    *it *= rv;
    ++it;
  }
  return *this;
}

template<class T>
CDimFeatureVector<T>& CDimFeatureVector<T>::operator/=(double rv) {
  iterator it = begin();
  while (it<end()) {
    *it /= rv;
    ++it;
  }
  return *this;
}

//--------------------------------
// shifts
//--------------------------------

template<class T>
void CDimFeatureVector<T>::shiftRight ( unsigned int first, unsigned int num ) {
  if (first >= size()) return;
  if (first+num > size()) num = (unsigned int)size()-first-1;
  if (num <= 1) return;
  T fe = (*this)[first+num-1];

  for (unsigned int i=num-1; i>0; --i)
    (*this)[first+i] = (*this)[first+i-1];
  (*this)[first] = fe;
}

template<class T>
void CDimFeatureVector<T>::shiftLeft  ( unsigned int first, unsigned int num ) {
  if (first >= size()) return;
  if (first+num > size()) num = (unsigned int)size()-first-1;
  if (num <= 1) return;
  T fe = (*this)[first];

  for (unsigned int i=0; i<num; ++i)
    (*this)[first+i] = (*this)[first+i+1];
  (*this)[first+num-1] = fe;
}


//--------------------------------
// prints
//--------------------------------

template<class T>
void CDimFeatureVector<T>::printf() {
 
  for (unsigned int i=0; i<size(); ++i) 
    std::cout << (*this)[i] << " ";
  std::cout << "\n";
}

#else
#pragma message(">>>>>  Feature vectors: simple (not STL)")

template <class T> 
class CDimFeatureVector
{
public:
  CDimFeatureVector( unsigned int dimensions );
  virtual ~CDimFeatureVector( );
  inline T& operator[](int x) { if (arr==NULL) return NULL; return arr[x]; }
  inline const T* data() { return arr; };
private:
  T *arr;
  unsigned int num_dims;
};

template<class T>
CDimFeatureVector<T>::CDimFeatureVector( unsigned int dimensions ) {
  num_dims = dimensions;
  try {  
     arr = new T[num_dims*sizeof(T)];
  } catch (...) { arr = NULL; };
}

template<class T>
CDimFeatureVector<T>::~CDimFeatureVector( ) {
  num_dims = 0;
  if (arr != NULL) delete []arr;
}

#endif // ifdef DIM_NOSTL_CONTAINER_VECTOR


//----------------------------------------------------------------------------
// Vector of features
// updated methods: erase clear resize
// restricted methods: pop_back push_back insert
//----------------------------------------------------------------------------

#ifndef DIM_NOSTL_CONTAINER_VECTOR

template <class T> 
class CDimVectorOfFeatures: public CDimVectorOwner< CDimFeatureVector<T> > 
{
public:
  typedef typename CDimVectorOwner< CDimFeatureVector<T> >::iterator iterator;

  CDimVectorOfFeatures( unsigned int dimensions ): num_dims(dimensions) {};
  CDimVectorOfFeatures( unsigned int dimensions, unsigned int size)
    : num_dims(dimensions) { resize(size); };
  virtual ~CDimVectorOfFeatures( ){ CDimVectorOwner< CDimFeatureVector<T> >::~CDimVectorOwner(); };
  void add( const T* feature_array, unsigned int size );
  void add( const CDimFeatureVector<T>* feature_vector );
  void fill( T v );
	void resize(size_type _Newsize);
	void resize(size_type _Newsize, T _Val);

  // I/O partial
  using CDimVectorOwner< CDimFeatureVector<T> >::fromFile;
  using CDimVectorOwner< CDimFeatureVector<T> >::toFile;

  int fromFile(std::ifstream *fs);
  int toFile(std::ofstream *fs);

  // data related
  // normalize each dimension by it's standard deviation
  void normalize_by_std();
  void normalize_by_std( unsigned int dim ); // only normalize dimension dim

protected:
  unsigned int num_dims;
};

template<class T>
void CDimVectorOfFeatures<T>::add( const T* fa, unsigned int size ) {
  unsigned int num = num_dims;
  if (size < num_dims) num = size;
  CDimFeatureVector<T> *nv=0;
  try { nv = new CDimFeatureVector<T> (num_dims); } catch (...) { return; };
  if (nv == 0) return;
  //memcpy(nv, feature_array, sizeof(T)*num_dims );
  for (unsigned int i=0; i<num; ++i)
    (*nv)[i] = fa[i];
  this->push_back( nv );
}

template<class T>
void CDimVectorOfFeatures<T>::add( const CDimFeatureVector<T>* feature_vector ) {
  if (feature_vector->size() != num_dims) return;
  CDimFeatureVector<T> *nv=0;
  try { nv = new CDimFeatureVector<T>(num_dims); } catch (...) { return; };
  if (nv == 0) return;
  nv->assign(feature_vector->begin(), feature_vector->end());
  this->push_back( nv );
}

template<class T>
void CDimVectorOfFeatures<T>::fill( T v ) {
  CDimVectorOfFeatures<T>::iterator it = begin();
  while ( it<end() ) {
    (*it)->fill(v);
    ++it;
  }
}

template<class T>
void CDimVectorOfFeatures<T>::resize(size_type _Newsize) {	
  T _Val = T();
  resize(_Newsize, _Val);
}

template<class T>
void CDimVectorOfFeatures<T>::resize(size_type _Newsize, T _Val) {	
  if (_Newsize == size()) return;
  _Val;
  if (_Newsize < size()) {
    CDimVectorOwner< CDimFeatureVector<T> >::resize(_Newsize);
  }
  else {
    unsigned int n = (unsigned int) _Newsize-size();
    CDimFeatureVector<T> fv(num_dims);
    for (unsigned int i=0; i<n; ++i)
      add( &fv );
  }
}

template<class T>
int CDimVectorOfFeatures<T>::fromFile(std::ifstream *fs) {
  this->clear();
  CDimFeatureVector<T> v(num_dims);
  while ( (v.fromFile(fs)>0) && (!fs->eof()) ) {
    this->add( &v );
  }
  return (int) this->size();
}


template<class T>
int CDimVectorOfFeatures<T>::toFile(std::ofstream *fs) {
  if (this->size() <= 0) return -1;  
  CDimVectorOfFeatures<T>::iterator it = begin();
  while ( (it<end()) && ((*it)->toFile(fs)>0) ) {
    ++it;
  }
  return (int) this->size();
}

template<class T>
void CDimVectorOfFeatures<T>::normalize_by_std() {
  if (this->size() <= 1) return; 
  CDimVectorOfFeatures<T>::iterator it;
  CDimFeatureVector<double> avgs( num_dims, 0 );
  CDimFeatureVector<double> stds( num_dims, 0 );
  unsigned int i;

  // get means
  it = begin();
  while ( it<end() ) {
    for (i=0; i<num_dims; ++i) avgs[i] += (*(*it))[i];
    ++it;
  }
  avgs /= (double) this->size();

  // get stds
  it = begin();
  while ( it<end() ) {
    for (i=0; i<num_dims; ++i) {
      register double v = (*(*it))[i] - avgs[i];
      stds[i] += v*v;
    }
    ++it;
  }
  for (i=0; i<num_dims; ++i)
    stds[i] = sqrt( stds[i]/(this->size() - 1.0) );

  // normalize
  it = begin();
  while ( it<end() ) {
    for (i=0; i<num_dims; ++i) if (stds[i] != 0.0) (*(*it))[i] /= (texture_type) stds[i];
    ++it;
  }
}

// this only normalizes dimension dim
template<class T>
void CDimVectorOfFeatures<T>::normalize_by_std( unsigned int dim ) {
  if (this->size() <= 1) return; 
  if (num_dims <= dim) return; 

  CDimVectorOfFeatures<T>::iterator it;
  double avgs = 0.0;
  double stds = 0.0;

  // get means
  it = begin();
  while ( it<end() ) {
    avgs += (*(*it))[dim];
    ++it;
  }
  avgs /= (double) this->size();

  // get stds
  it = begin();
  while ( it<end() ) {
    register double v = (*(*it))[dim] - avgs;
    stds += v*v;
    ++it;
  }
  stds = sqrt( stds/(this->size() - 1.0) );

  // normalize
  it = begin();
  while ( it<end() ) {
    if (stds != 0.0) (*(*it))[dim] /= (texture_type) stds;
    ++it;
  }
}

#else // ifndef DIM_NOSTL_CONTAINER_VECTOR

template <class T> 
class CDimVectorOfFeatures: public std::vector <T*> 
{
public:
  CDimVectorOfFeatures( unsigned int dimensions ): num_dims(dimensions) {};
  void add( const T* feature_array, unsigned int size );
  void add( const std::vector <T*> feature_vector );

  // make sure to remove all stuff created in memory
  iterator erase( iterator Where ) {
    if (Where != NULL) {
      delete[] *Where;
      *Where = NULL;    
    }
    return std::vector<T*>::erase( Where );
  }
  
  iterator erase( iterator First, iterator Last ) {
    while(First != Last) {
      delete[] *First;
      *First = NULL;
      ++First;
    }    
    return std::vector<T*>::erase( First, Last );
  };

  void clear() {
		erase(begin(), end());
    return std::vector<T*>::clear();
	}

	void resize(size_type _Newsize) {	
		resize(_Newsize, _Ty());
	}

	void resize(size_type _Newsize, T _Val) {	
    if (_Newsize < size())
			erase(begin() + _Newsize, end());
    std::vector<T*>::resize(_Newsize, _Val);
	}


protected:
  unsigned int num_dims;
  
  // restrict access to STL container methods in order to enforce the new interface
  using std::vector < CDimFeatureVector<T>* >::pop_back;
  using std::vector < CDimFeatureVector<T>* >::push_back;
  using std::vector < CDimFeatureVector<T>* >::insert;
};

template<class T>
void CDimVectorOfFeatures<T>::add( const T* feature_array, unsigned int size ) {
  if (size < num_dims) return;
  T* nv;
  try { nv = new T[num_dims]; } catch (...) { return; };
  if (nv == 0) return;
  memcpy(nv, feature_array, sizeof(T)*num_dims );

  this->push_back( nv );
}

template<class T>
void CDimVectorOfFeatures<T>::add( const std::vector <T*> feature_vector ) {
  if (feature_vector.size() < num_dims) return;

  try {  
     T* nv = new T[num_dims];
  } catch (...) { return };
  if (nv == 0) return;
  for (int i=0; i<num_dims; ++i)
    nv[i] = feature_vector[i];

  this->push_back( nv );
}

#endif // ifdef DIM_NOSTL_CONTAINER_VECTOR

#endif // DIM_FEATUREVECTORS
