/*****************************************************************************
 descriptors for textures, tiles and images

 DEFINITION

 Author: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>
 Copyright (c) 2006 Vision Research Lab, UCSB <http://vision.ece.ucsb.edu>

 History:
   04/19/2006 16:20 - First creation

 Ver : 1
*****************************************************************************/

#ifndef DIM_TEXTURE_DESCRIPTORS
#define DIM_TEXTURE_DESCRIPTORS

#include <string>
#include <vector>
#include <deque>
#include <iostream>
#include <fstream>

#include <xtypes.h>
#include <xvectors.h>

// windows: use secure C libraries with VS2005 or higher
#if ( defined(_MSC_VER) && (_MSC_VER >= 1400) )
  #define HAVE_SECURE_C
  #pragma message(">>>>>  textureDescriptors: using secure c libraries")
#endif 


//The feature vectors have dimension 60.  
//They are the mean and standard deviations of 30 Gabors filters (5 scales and 6 orientations).  
//The exact ordering of the elements is not known.
typedef	float texture_type;
const int TEXTURE_DIMS = 60;

// layer GT data: 0-BG 1-ILM 2-GCL 3-IPL 4-INL 5-OPL 6-ONL 7-IS 8-OS 8-RPE 10-M/L
const int RET_LAYER_BG  = 0;  // background
const int RET_LAYER_ILM = 1;   
const int RET_LAYER_GCL = 2;
const int RET_LAYER_IPL = 3;  // inner plexiform layer
const int RET_LAYER_INL = 4;  // inner nuclear layer
const int RET_LAYER_OPL = 5;  // outer plexiform layer
const int RET_LAYER_ONL = 6;  // outer nuclear layer
const int RET_LAYER_IS  = 7;  
const int RET_LAYER_OS  = 8;
const int RET_LAYER_RPE = 9;
const int RET_LAYER_M_L = 10;
 
//----------------------------------------------------------------------------
// BASES
//----------------------------------------------------------------------------

class CDimDescriptorBase: public CDimListIO
{
public:
  virtual CDimDescriptorBase* clone() const  { return new CDimDescriptorBase(*this); };
  virtual ~CDimDescriptorBase() {};
  virtual void assign(const CDimDescriptorBase &d) { d; };
};

template <class T> 
class CDimItemListBase: public CDimDequeOwner<T> 
{
public:
  virtual ~CDimItemListBase( ) { CDimDequeOwner<T>::~CDimDequeOwner(); };

  // I/O
  using CDimDequeOwner<T>::fromFile;
  using CDimDequeOwner<T>::toFile;
  int fromFile(std::ifstream *fs);
  int toFile(std::ofstream *fs);
};

template<class T>
int CDimItemListBase<T>::fromFile(std::ifstream *fs) {
  T td;
  this->clear();
  while ( (td.fromFile(fs)>0) || (!fs->eof()) ) {
    this->add( td );
  }
  return (int) this->size();
}

template<class T>
int CDimItemListBase<T>::toFile(std::ofstream *fs) {
  if (this->size() <= 0) return -1;  

  unsigned int i = 0;
  while (i < size()) {
    if (at(i)->toFile(fs) == -1) break;
    ++i;
  }
  return (int) this->size();
}

//----------------------------------------------------------------------------
// tiles
//----------------------------------------------------------------------------

class CDimTileDescriptor: public CDimDescriptorBase
{
public:
  // info
  int            image_id;           // image ID, handle to 3D image list
  unsigned int   tile_id;            // tile ID, computed in rows->columns order
  unsigned int   tile_size_pixels;   // tile width/height in pixels, it's square
  unsigned int   position_in_img[5]; // XYZCT XY-of the top left corner
  // descriptors
  texture_type   texture[TEXTURE_DIMS]; // 60 dimensional texture descriptor
  // assignements
  int            class_assigned;      // assignement to a class
  int            class_groundtruth;   // handmade assignement to a class
  unsigned char  layers_assigned[12]; // 12 layers - 3 int32
  // layer GT data: 0-BG 1-ILM 2-GCL 3-IPL 4-INL 5-OPL 6-ONL 7-IS 8-OS 8-RPE 10-M/L
  unsigned char  layers_gt[12];       // 12 layers - 3 int32

  //--------------------------
  // methods
  //--------------------------
  // int32  *1   image_id
  // uint32 *1   tile_id
  // uint32 *1   tile_size_pixels
  // uin32  *3   position_in_img - only first 3 positions XYZ
  // int    *1   class_assigned
  // int    *1   class_groundtruth
  // char   *12  layers_assigned - percentage of each layer in the given tile
  // char   *12  layers_gt - ground truth of percentage of each layer in the given tile 
  // float  *60  texture

  virtual unsigned int   bufferSize();                    // size needed to store the buffer 
  virtual unsigned char* toBuffer();                      // creates in memory a buffer with tile descriptor
  virtual int            fromBuffer(const unsigned char* buf);  // loads descriptor from the buffer

  // IO
  int fromFile(std::ifstream *fs);
  int toFile(std::ofstream *fs);

  CDimTileDescriptor();
  virtual ~CDimTileDescriptor();

  void assign(const CDimTileDescriptor &d);
private:

};

//----------------------------------------------------------------------------
// images
//----------------------------------------------------------------------------

class CDimImageItem: public CDimDescriptorBase
{
public:
  // info
  int           image_id;           // 3D image ID, handle to image list
  std::string   image_name;         // image file name
  unsigned int  image_size[5];      // image size WHZCT
  double        pixel_size[3];      // XYZ

  // IO
  int fromFile(std::ifstream *fs);
  int toFile(std::ofstream *fs);

  CDimImageItem();
  ~CDimImageItem();
};

class CDimMapItem: public CDimDescriptorBase
{
public:
  // info
  int           image_plane_id;     // image ID in old by plane list
  int           image_5d_id;        // image ID in the new 3D list

  // IO
  int fromFile(std::ifstream *fs);
  int toFile(std::ofstream *fs);

  CDimMapItem();
};

//----------------------------------------------------------------------------
// lists
//----------------------------------------------------------------------------

// header: TF30 - 4 bytes
class CDimTileList: public CDimItemListBase <CDimTileDescriptor> {
public:
  // I/O
  int readHeader( std::ifstream *fs ); 
  int writeHeader( std::ofstream *fs );

  // a very slow function to retrieve tile position
  int getTileIndex(int img_id, unsigned int x, unsigned int y, unsigned int z);
  
  // return the number of clusters assigned to the data points
  // this is a bit cheating, we'll get the maximum value of class_assigned
  // presumed: all the classes are assigned consequtively starting with 0
  int getK();

  
  // these functions create an average texture vector
  
  // gt_class - id of the ground truth class of interest
  // min_gt_prob - minimum probability of GT class, if 0 then the class_groundtruth value will be used instead
  // min_gt_prob: 0 - FULL, 90 - CLEAR, 100 - PURE
  CDimTileDescriptor getMeanTextureGT( int gt_class, int min_gt_prob );
  
  // as_class - id of the assigned class of interest
  CDimTileDescriptor getMeanTextureAssigned( int as_class );
};

class CDimTileListExt: public CDimTileList {
public:
  int getMaxPosInCol(int img_id);
  int getMaxPosInRow(int img_id);
};


class CDimImageList: public CDimItemListBase <CDimImageItem> { 
public:
  int readHeader( std::ifstream *fs ); 
  int writeHeader( std::ofstream *fs ); 
};

class CDimMapList: public CDimItemListBase <CDimMapItem> { 
public:

};

#endif // DIM_TEXTURE_DESCRIPTORS
